<?php
/**
 * Abstract class that all Registrar Modules must extend
 *
 * @package blesta
 * @subpackage blesta.components.modules
 * @copyright Copyright (c) 2021, Phillips Data, Inc.
 * @license http://www.blesta.com/license/ The Blesta License Agreement
 * @link http://www.blesta.com/ Blesta
 */
abstract class RegistrarModule extends Module
{
    /**
     * Used to determine whether this module supports DNS Management
     *
     * @return bool True to say that this module uses the dns_management configurable option, false otherwise
     */
    public function supportsDnsManagement()
    {
        return false;
    }

    /**
     * Used to determine whether this module supports Email Forwarding
     *
     * @return bool True to say that this module uses the email_forwarding configurable option, false otherwise
     */
    public function supportsEmailForwarding()
    {
        return false;
    }

    /**
     * Used to determine whether this module supports ID Protection
     *
     * @return bool True to say that this module uses the id_protection configurable option, false otherwise
     */
    public function supportsIdProtection()
    {
        return false;
    }

    /**
     * Used to determine whether this module supports EPP Code
     *
     * @return bool True to say that this module uses the epp_code configurable option, false otherwise
     */
    public function supportsEppCode()
    {
        return false;
    }

    /**
     * Checks if this registrar supports a given feature
     *
     * @param string $feature_name The name of the feature for which to check support (e.g. id_protection)
     */
    public function supportsFeature($feature_name)
    {
        $method_name = 'supports' . str_replace('_', '', ucwords($feature_name, '_'));
        return method_exists($this, $method_name) && $this->{$method_name}();
    }

    /**
     * Returns the type of this module
     *
     * @return string The type of this module
     */
    public function getType()
    {
        if (isset($this->config->type)) {
            return $this->config->type;
        }

        return 'registrar';
    }

    /**
     * Verifies that the provided domain name is available
     *
     * @param string $domain The domain to lookup
     * @param int $module_row_id The ID of the module row to fetch for the current module
     * @return bool True if the domain is available, false otherwise
     */
    public function checkAvailability($domain, $module_row_id = null)
    {
        return true;
    }

    /**
     * Gets the domain name from the given service
     *
     * @param stdClass $service The service from which to extract the domain name
     * @return string The domain name associated with the service
     * @see Services::get()
     */
    public function getServiceDomain($service)
    {
        if (isset($service->fields)) {
            foreach ($service->fields as $service_field) {
                if ($service_field->key == 'domain') {
                    return $service_field->value;
                }
            }
        }

        return $this->getServiceName($service);
    }

    /**
     * Gets the domain expiration date
     *
     * @param stdClass $service The service belonging to the domain to lookup
     * @param string $format The format to return the expiration date in
     * @return string The domain expiration date in UTC time in the given format
     * @see Services::get()
     */
    public function getExpirationDate($service, $format = 'Y-m-d H:i:s')
    {
        if (isset($service->date_renews)) {
            return $service->date_renews;
        }

        return null;
    }

    /**
     * Get a list of the TLDs supported by the registrar module
     *
     * @param int $module_row_id The ID of the module row to fetch for the current module
     * @return array A list of all TLDs supported by the registrar module
     */
    public function getTlds($module_row_id = null)
    {
        return [];
    }
}
